#ident "$Id: abcprintd.c,v 1.1 2004/09/09 08:13:32 hpa Exp $"
/* ----------------------------------------------------------------------- *
 *   
 *   Copyright 2004 H. Peter Anvin - All Rights Reserved
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
 *   Bostom MA 02111-1307, USA; either version 2 of the License, or
 *   (at your option) any later version; incorporated herein by reference.
 *
 * ----------------------------------------------------------------------- */

/*
 * abcprintd.c
 *
 * Read a serial port; send data to lpr when receiving BREAK
 *
 * Usage: abcprintd port [options to lpr]
 */

#include <stdio.h>
#include <unistd.h>
#include <termios.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <fcntl.h>

static int lpr_argc;
static const char **lpr_argv;

static void print_setup(FILE **tfp)
{
  FILE *tf = *tfp;
  pid_t f;

  if ( tf ) {
    fflush(tf);
    rewind(tf);

    f = fork();

    if ( f < 0 ) {
      perror("fork");
      exit(1);
    } else if ( f == 0 ) {
      dup2(fileno(tf), STDIN_FILENO);
      fclose(tf);
      execvp(lpr_argv[0], (char **)lpr_argv);
      _exit(255);
    } else {
      fclose(tf);
      while ( waitpid(f, NULL, 0) != f );
    }
  }

  *tfp = tmpfile();
  if ( !*tfp ) {
    perror("tmpfile");
    exit(1);
  }
}

static int open_port(const char *path)
{
  struct termios tio;
  int fd;
  int flags;

  if ( (fd = open(path, O_RDWR|O_NONBLOCK)) < 0 )
    return -1;

  if ( tcgetattr(fd, &tio) )
    return -1;

  tio.c_iflag &= ~(BRKINT|IGNBRK|IGNPAR|INPCK|ISTRIP|ICRNL|INLCR|IGNCR|
		   IXON|IXOFF);
  tio.c_iflag |= PARMRK;

  tio.c_cflag &= ~(CSIZE|CSTOPB|HUPCL|PARENB|PARODD);
  tio.c_cflag |= CLOCAL|CREAD|CS8;

  tio.c_oflag &= ~OPOST;

  tio.c_lflag &= ~(ECHO|ECHOE|ECHOK|ECHONL|ICANON|ISIG|NOFLSH|TOSTOP|IEXTEN);

  tio.c_cc[VMIN]  = 64;
  tio.c_cc[VTIME] = 1;

  cfsetispeed(&tio, B115200);
  cfsetospeed(&tio, B115200);

  tcflush(fd, TCIFLUSH);

  if ( tcsetattr(fd, TCSANOW, &tio) )
    return -1;

  if ( (flags = fcntl(fd, F_GETFL)) == -1 )
    return -1;

  flags &= ~O_NONBLOCK;

  if ( fcntl(fd, F_SETFL, flags) )
    return -1;

  return fd;
}

static void output(int c, FILE *tf)
{
  /* This is a separate function so we can put any needed data modifications
     in this file */
  fwrite(&c, 1, 1, tf);
}

#define BUF_SIZE 4096

int main(int argc, char *argv[])
{
  int fd;
  FILE *tf = NULL;
  enum {
    st_normal,			/* Normal operation */
    st_ff,			/* 0xFF received */
    st_ff00,			/* 0xFF 0x00 received */
  } state;
  char ibuf[BUF_SIZE], c;
  int i, b;
  static const char *lpr_default_argv[] = {
    "enscript", "-B", "-Z", "-q", "-X", "asciise",
    "--non-printable-format=space",
    NULL
  };

  if ( argc < 2 ) {
    fprintf(stderr, "Usage: %s port [lpr_program [options...]]\n", argv[0]);
    exit(1);
  }

  if ( (fd = open_port(argv[1])) < 0 ) {
    perror(argv[1]);
    exit(1);
  }

  if ( argc == 2 ) {
    lpr_argc = sizeof(lpr_default_argv)/sizeof(const char *) - 1;
    lpr_argv = lpr_default_argv;
  } else {
    lpr_argc = argc-2;
    if ( !(lpr_argv = malloc(sizeof(char *) * (lpr_argc+1))) ) {
      perror("malloc");
    exit(1);
    }
    memcpy(lpr_argv, argv+2, sizeof(char *) * (argc-2));
  }

  daemon(0,0);

  print_setup(&tf);

  state = st_normal;

  for ( ;; ) {
    if ( (b = read(fd, ibuf, sizeof ibuf)) < 1 ) {
      perror(argv[1]);
      exit(1);
    }

    for ( i = 0 ; i < b ; i++ ) {
      c = ibuf[i];

      switch ( state ) {
      case st_normal:
	if ( c == '\xff' )
	  state = st_ff;
	else
	  output(c, tf);
	break;

      case st_ff:
	if ( c != 0 ) {
	  output(c, tf);
	  state = st_normal;
	} else {
	  state = st_ff00;
	}
	break;

      case st_ff00:
	if ( c == 0 )
	  print_setup(&tf);	/* BREAK received */
	state = st_normal;
	break;
      }
    }
  }
}

	  
	
  
